#include "rpg.h"

#define SMALL_DELAY 1000000L
#define LARGE_DELAY 2000000L 

char msg_line[5][26];
byte color[5];

void sendText(char *message, byte clr)
{
  int x;
  for (x = 0; x < 4; x++)
  {
    strcpy(msg_line[x], msg_line[x + 1]);
    color[x] = color[x + 1];
  }
  strcpy(msg_line[4], message);
  color[4] = clr;

  window(60, 140, 200, 40);
  for (x = 0; x < 5; x++)
    blitString(60, 140 + x * 8, color[x], msg_line[x]);
}

//////////////////////////////////////////////////////////////////////////////

void clearText(void)
{
  sendText("", 0);
  sendText("", 0);
  sendText("", 0);
  sendText("", 0);
  sendText("", 0);
}

//////////////////////////////////////////////////////////////////////////////

monsterType mdata;

void printStatus(void)
{
  char string[12];
  window(10, 10, 90, 110);         // Stat window

  blitString(10, 10, 15, "HP");
  sprintf(string, "%d", player.HP);
  blitString(10, 18, 15, string);
  blitChar(10 + 4 * 8, 18, 15, '/');
  sprintf(string, "%d", player.maxHP);
  blitString(10 + 5 * 8, 18, 15, string);

  blitString(10, 34, 15, "MP");
  sprintf(string, "%d", player.MP);
  blitString(10, 42, 15, string);
  blitChar(10 + 4 * 8, 42, 15, '/');
  sprintf(string, "%d", player.maxMP);
  blitString(10 + 5 * 8, 42, 15, string);
}

//////////////////////////////////////////////////////////////////////////////

void showCommands(void)
{
  window(220, 10, 84, 72);        // Command window
  blitString(240, 14, 15, "Attack");
  blitString(240, 30, 15, "Magic");
  blitString(240, 46, 15, "Item");
  blitString(240, 62, 15, "Run");
}

//////////////////////////////////////////////////////////////////////////////

int getCommand(void)
{
  showCommands();

  blitImageOpaque(220, 10, &arrow);

  int choice = 0, oldchoice;
  int done = 0;
  while (!done)
  {
    oldchoice = choice;

    while (keyPressed()) {}

    int key = 0;
    while (!key)
    {
      if (keyTable[DOWNARROW]) { if (++choice == 4) choice = 0;  key = 1; }
      if (keyTable[UPARROW])   { if (--choice == -1) choice = 3; key = 1; }
      if (keyTable[ENTER])     { done = 1;                       key = 1; }
    }

    if (oldchoice != choice)
    {
      clearArea(220, 10 + oldchoice * 16, 16, 16, 0);
      blitImageOpaque(220, 10 + choice * 16, &arrow);
      PlaySound(SOUND_CHOICE);
    }
  }

  clearArea(220, 10 + choice * 16, 16, 16, 0);
  return choice;
}

//////////////////////////////////////////////////////////////////////////////

void monsterMove(void)
{
  char string[26];

  int damage;
  int minDamage, maxDamage;
  minDamage = mdata.attack * .8;
  if (minDamage == mdata.attack) minDamage--;
  if (minDamage < 1) minDamage = 1;

  maxDamage = mdata.attack * 1.2;
  if (maxDamage == mdata.attack) maxDamage++;
  if (maxDamage > 9999) maxDamage = 9999;

  damage = random(maxDamage - minDamage) + minDamage;

  damage *= (100.0 - (float)player.defense()) / 100.0;
  if (damage < 1) damage = 1;
  player.HP -= damage;

  if (player.HP < 0) player.HP = 0;

  sprintf(string, "%s does %d", mdata.name, damage);
  sendText("", 0);
  sendText(string, 4);
  sendText("damage.", 4);
  HDelay(SMALL_DELAY);

  printStatus();
}

//////////////////////////////////////////////////////////////////////////////

int getSpell(void)
{
  int spell = 0;

  int x;
  for (x = 0; x < 8; x++)
    if (player.spells[x]) spell = 1;
  if (spell == 0)
    return -2;

  int curchoice = 0, oldchoice = 0;
  blitImage(220, 10, &arrow);

  int done = 0;
  while (!done)
  {
    if (oldchoice != curchoice) PlaySound(SOUND_CHOICE);
    oldchoice = curchoice;

    while (keyPressed()) {}
    int key = 0;
    while (!key)
    {
      if (keyTable[UPARROW])
      {
        key = 1;
        do { if (--curchoice == -1) curchoice = 7; }
        while (!player.spells[curchoice]);
      }
      if (keyTable[DOWNARROW])
      {
        key = 1;
        do { if (++curchoice ==  8) curchoice = 0; }
        while (!player.spells[curchoice]);
      }
      if (keyTable[ENTER])     { key = 1; done = 1; }
      if (keyTable[ESC])       { key = 1; done = 2; }
    }

    clearArea(220, 10 + oldchoice * 8, 16, 16);
    blitImage(220, 10 + curchoice * 8, &arrow);
  }

  clearArea(220, 10 + curchoice * 8, 16, 16);

  if (done == 2) return -1; else return curchoice;
}

//////////////////////////////////////////////////////////////////////////////

#define items(x)    gamedata.items[x]
#define itemData(x) gamedata.itemData[x]
#define weapons(x)  gamedata.weapons[x]
#define shields(x)  gamedata.shields[x]
#define helmets(x)  gamedata.helmets[x]
#define mails(x)    gamedata.mails[x]

void showitems(int topitem)
{
  char string[80];

  window(220, 10, 84, 72);        // Command window
  int x;
  for (x = topitem; x < topitem + 8; x++)
  {
    int itemnum = items(x).num;
    string[0] = 0;
    switch (items(x).type)
    {
      case BLANK:  break;
      case ITEM:   strcpy(string, itemData(itemnum).name); break;
      case WEAPON: strcpy(string, weapons(itemnum).name);  break;
      case SHIELD: strcpy(string, shields(itemnum).name);  break;
      case HELMET: strcpy(string, helmets(itemnum).name);  break;
      case MAIL:   strcpy(string, mails(itemnum).name);    break;
    }
    string[8] = 0;
    blitString(240, (x - topitem) * 8 + 14, 15, string);
  }
}

//  //  //  //  //  //  //  //  //  //  //  //  //  //  //  //  //  //  //  //

void useitem(void)
{
  showitems(0);
  blitImage(220, 10, &arrow);

  int topitem = 0, curitem = 0;

  int done = 0;
  while (!done)
  {
    int exec = 0;

    while (keyPressed()) {}
    int key = 0;
    while (!key)
    {
      if (keyTable[UPARROW])   { key = 1; if (--curitem == topitem - 1)
                                 if (--topitem == -1)
                                   { topitem++; curitem++; }
                                 PlaySound(SOUND_CHOICE);
                               }
      if (keyTable[DOWNARROW]) { key = 1; if (++curitem == topitem + 8)
                                 if (++topitem == NUM_ITEMS)
                                   { topitem--; curitem--; }
                                 PlaySound(SOUND_CHOICE);
                               }
      if (keyTable[ESC])       { key = 1; done = 1; }
      if (keyTable[ENTER])     { key = 1; exec = 1; }
    }

    if (exec || done) PlaySound(SOUND_SELECT);

    if (exec)
    {
      if (items(curitem).type == ITEM)
      {
        int itemnum = items(curitem).num;
        player.HP += itemData(itemnum).HPup;
        if (player.HP > player.maxHP) player.HP = player.maxHP;
        player.MP += itemData(itemnum).MPup;
        if (player.MP > player.maxMP) player.MP = player.maxMP;
        player.status |= itemData(itemnum).status;
        items(curitem).type = BLANK;
        printStatus();
      }
    }

    showitems(topitem);
    blitImage(220, (curitem - topitem) * 8 + 10, &arrow);
  }
}

//////////////////////////////////////////////////////////////////////////////

#define OK            0
#define MONSTER_DIED  1
#define RUN           2
#define PLAYER_CANCEL 3

int playerMove(void)
{
  char string[26];
  int minDamage, maxDamage;
  int damage;

  int command;
  command = getCommand();
  switch (command)
  {
    case 0:
      PlaySound(SOUND_P_ATTACK);
      minDamage = player.attack() * .8;
      if (minDamage == player.attack()) minDamage--;
      maxDamage = player.attack() * 1.2;
      if (maxDamage == player.attack()) maxDamage++;
      damage = random(maxDamage - minDamage) + minDamage;
      sprintf(string, "You do %d damage.", damage);
      sendText("", 0);
      sendText(string, 13);
      HDelay(SMALL_DELAY);
      clearText();
      mdata.HP -= damage;
      if (mdata.HP < 0) mdata.HP = 0;
      break;

    case 1:
      window(220, 10, 84, 72);        // Command window
      int x;
      for (x = 0; x < 8; x++)       // Assumes there are 8 spells
        if (player.spells[x])
          blitString(240, 14 + x * 8, 15, gamedata.spells[x].name);

      int spell;
      switch (spell = getSpell())
      {
        case -2: sendText("No spells", 12); return PLAYER_CANCEL;
        case -1: return PLAYER_CANCEL;
        default:
          if (gamedata.spells[spell].cost > player.MP)
            { sendText("Not enough MP", 12); return PLAYER_CANCEL; }
          PlaySound(SOUND_SPELL);
          player.MP -= gamedata.spells[spell].cost;
        
          if (gamedata.spells[spell].effect == DAMAGE)
          {
            int damage = gamedata.spells[spell].factor * player.wisdom;
            damage *= (100.0 - (float)mdata.mdef) / 100.0;
            sprintf(string, "%s did %d damage.",
                    gamedata.spells[spell].name, damage);
            sendText("", 0);
            sendText(string, 4);
            HDelay(SMALL_DELAY);

            mdata.HP -= damage;
            if (mdata.HP < 0) mdata.HP = 0;
          }
          else
          if (gamedata.spells[spell].effect == HEALING)
          {
            int healing = gamedata.spells[spell].factor * player.wisdom;
            sprintf(string, "%s did %d healing.",
                    gamedata.spells[spell].name, healing);
            sendText("", 0);
            sendText(string, 10);
            HDelay(SMALL_DELAY);

            player.HP += healing;
            if (player.HP > player.maxHP) player.HP = player.maxHP;
          }
          else
          if (gamedata.spells[spell].effect == MAKESTATUS)
          {
            int percent = ((float)player.wisdom / 50.0) * 100.0;
            if (percent > 99) percent = 90;
            if (random(100) > percent)
            {
              sendText("", 0);
              sendText("Failed", 12);
              HDelay(SMALL_DELAY);
            }
            else
            {
              if (gamedata.spells[spell].target == TGT_PLAYER)
                player.status |= gamedata.spells[spell].factor;
              else
                mdata.status |= gamedata.spells[spell].factor;
            }
          }
          else
          if (gamedata.spells[spell].effect == RMVSTATUS)
          {
            int percent = ((float)player.wisdom / 50.0) * 100.0;
            if (percent > 99) percent = 90;
            if (random(100) > percent)
            {
              sendText("", 0);
              sendText("Failed", 12);
              HDelay(SMALL_DELAY);
            }
            else
            {
              if (gamedata.spells[spell].target == TGT_PLAYER)
                player.status &= ~gamedata.spells[spell].factor;
              else
                mdata.status &= ~gamedata.spells[spell].factor;
            }
          }
      }
      break;

    case 2:
      PlaySound(SOUND_SELECT);
      useitem();
      return PLAYER_CANCEL;

    case 3:
      PlaySound(SOUND_RUN);
      if (random(mdata.speed + 2) > random(player.speed + 2))
        { sendText("You failed to run away!", 4); return OK; }
      else
        { sendText("You ran away!", 9); return RUN; }
  }
  printStatus();

  if (mdata.HP < 1)
  {
    char string[26];

    PlaySound(SOUND_VICTORY);
    sendText("The monster died.", 10);

    player.gold += mdata.gold;
    if (player.gold > 99999) player.gold = 99999;
    sprintf(string, "You got %d gold.", mdata.gold);
    sendText(string, 14);

    if (player.level < MAX_LEVEL - 1) player.exp += mdata.exp;
    sprintf(string, "You got %d exp.", mdata.exp);
    sendText(string, 7);

    int done = 0;
    while (!done)
    {
      if (player.level < MAX_LEVEL - 1)
      {
        if (player.exp > gamedata.levelTable[player.level + 1])
        {
          sendText("", 0);
          sendText("You went up a level!", 9);
          HDelay(LARGE_DELAY);
          PlaySound(SOUND_LEVELUP);
          player.level++;

          if (random(2)) player.strength++;
          if (random(2)) player.speed++;
          if (random(2)) player.wisdom++;

          int statUp;

          statUp = random(player.maxHP) / 3;
          if (statUp < 1) statUp = 1;
          player.maxHP += statUp;

          statUp = random(player.maxMP) / 4;
          if (statUp < 1) statUp = 1;
          player.maxMP += statUp;

          int index;
          for (index = 0; index < 8; index++)
            if ((player.level + 1) >= gamedata.spells[index].level)
              if (player.spells[index] == 0)
              {
                player.spells[index] = 1;
                sendText("You learned a spell!", 13);
                sendText(gamedata.spells[index].name, 15);
              }

        } else done = 1;
      } else done = 1;
    }

    return MONSTER_DIED;
  }

  return OK;
}

//////////////////////////////////////////////////////////////////////////////

void battle(int type, int backnum)
{
  battleType far *battleData;
  battleData = (battleType far *)farmalloc(sizeof(battleType));
  if (battleData == NULL)
    errorAbort("Error: Not enough memory\n\r");

  int monster;
  FILE *handle;
  int i;

  handle = fopen(BTYPE_FILE, "rb");
  if (handle == NULL)
  {
    farfree(battleData);
    errorAbort("Error: Could not open battle file\n\r");
  }
  fseek(handle, type * sizeof(battleType), SEEK_SET);
  for (i = 0; i < sizeof(battleType); i++)
    ((byte far *)battleData)[i] = fgetc(handle);
  fclose(handle);

  if (random(100) >= battleData->percent)
    { farfree(battleData); return; }

  if (battleData->numMonsters < 1)
    { farfree(battleData); return; }

  PlaySound(SOUND_BATTLE);

  monster = battleData->monster[random(battleData->numMonsters)];

  farfree(battleData);

  memset(msg_line, 0, 26 * 5);

  handle = fopen(MONSTER_FILE, "rb");
  if (handle == NULL)
    errorAbort("Error: Could not open monster file\n\r");
  fseek(handle, monster * sizeof(monsterType), SEEK_SET);
  fread(&mdata, 1, sizeof(monsterType), handle);
  fclose(handle);

  byte far *mpic = (byte far *)farmalloc(80 * 80);
  if (mpic == NULL)
    errorAbort("Error: Not enough memory\n\r");

  byte far *mback = (byte far *)farmalloc(80 * 80);
  if (mback == NULL)
    { farfree(mpic); errorAbort("Error: Not enough memory\n\r"); }

  handle = fopen(MONSTERIMAGE_FILE, "rb");
  if (handle == NULL)
    { farfree(mpic); farfree(mback);
      errorAbort("Error: Could not open monster image file\n\r"); }
  fseek(handle, (long)monster * 80L * 80L, SEEK_SET);
  for (i = 0; i < 80 * 80; i++)
    mpic[i] = fgetc(handle);
  fclose(handle);

  handle = fopen(BATTLEBACK_FILE, "rb");
  if (handle == NULL)
    { farfree(mpic); farfree(mback);
      errorAbort("Error: Could not open battle background file\n\r"); }
  fseek(handle, (long)backnum * 80L * 80L, SEEK_SET);
  for (i = 0; i < 80 * 80; i++)
    mback[i] = fgetc(handle);
  fclose(handle);

  mdata.HP = mdata.maxHP;

  byte far *oldbuf = curBuf;
  curBuf = screen;

  window(112, 20, 96, 10);         // Monster name window
  blitString(112, 21, 15, mdata.name);

  window(120, 40, 80, 80);         // Monster picture
  int x, y;

  for (x = 0; x < 80; x++)
    for (y = 0; y < 80; y++)
      curBuf[(y + 40) * 320 + (x + 120)] = mback[y * 80 + x];

  int pixel;
  for (x = 0; x < 80; x++)
    for (y = 0; y < 80; y++)
    {
      pixel = mpic[y * 80 + x];
      if (pixel) curBuf[(y + 40) * 320 + (x + 120)] = pixel;
    }

  farfree(mpic);
  farfree(mback);

  printStatus();

  sendText("You have been attacked!", 15);

  int done = 0;

  enum { MONSTER, PLAYER } curTurn;

  if (random(mdata.speed + 2) > random(player.speed + 2))
    { curTurn = MONSTER; sendText("Monster moves!", 46); }
  else
    { curTurn = PLAYER;  sendText("Move first!", 79); }
  HDelay(SMALL_DELAY);
  clearText();

  while (!done)
  {
    if (curTurn == MONSTER)
    {
      monsterMove();

      if (player.status & STATUS_POISON)
      {
        char string[80];

        int damage = (float)player.maxHP / 4.0;
        sprintf(string, "Poison does %d", damage);
        sendText("", 0);
        sendText(string, 4);
        sendText("damage.", 4);
        HDelay(SMALL_DELAY);

        player.HP -= damage;
        if (player.HP < 0) player.HP = 0;
      }

      if (player.HP < 1)
      {
        PlaySound(SOUND_DEATH);
        sendText("You have died.", 8);
        HDelay(LARGE_DELAY);
        done = 1;
      }
    }

    if (curTurn == PLAYER)
    {
      int event;
      event = playerMove();

      if (event == MONSTER_DIED)  done = 1;
      if (event == RUN)           done = 1;
      if (event == PLAYER_CANCEL) curTurn = MONSTER;
    }

    if (curTurn == PLAYER)
    {
      if (player.status & STATUS_FAST)
        { sendText("You are fast!", 14); if (random(2)) curTurn = MONSTER; }
      else curTurn = MONSTER;
    }
    else
    {
      if (player.status & STATUS_SLOW)
        { sendText("You are slow!", 14); if (random(2)) curTurn = PLAYER; }
      else curTurn = PLAYER;
    }
  }

  while (keyPressed()) {}
  while (!keyPressed()) {}

  curBuf = oldbuf;
  blitScreen(oldbuf);
}
