#include "rpg.h"

byte far *mainbuf;
paletteType defaultPalette;
paletteType gamePalette;
imageType arrow;
byte far *map;
byte far *battleMap;
tileType far *tiles;
byte walkable[128];
tileType far *sprites;

subAreaType far *subAreaData;
messageType messages[NUM_MESSAGES];
gameDataType gamedata;
playerType player;

configType gameconfig;

int mapx, mapy;
int direction;
int gamestatus = NORMAL;

int sb_exist;

char *sound_key[NUMSOUNDS] =
{
  "CHOICE", "SELECT", "LEVELUP", "WORDBEEP", "CHEST", "INN", "NEWGAME",
  "SUBAREA", "BATTLE", "DEATH", "ENDING", "VICTORY", "SPELL", "P_ATTACK",
  "M_ATTACK", "RUN"
};
int baseio, irq, dma, dma16;
SOUND *sounds[NUMSOUNDS];

enum
{
  ALLOC_MEM = 1,
  LOAD_DATA,
  INIT_SOUND,
  SET_PALETTE
} errorCode;

//////////////////////////////////////////////////////////////////////////////

void errorAbort(char *message)
{
  Close();
  cputs(message);
  exit(1);
}

//////////////////////////////////////////////////////////////////////////////

void soundexitproc(void)
{
  int i;
  for (i = 0; i < NUMSOUNDS; i++)
    if (sounds[i] != NULL)
      free_sound(sounds + i);

#ifdef SHAREDEMB
  shutdown_sharing();
#endif
}

//////////////////////////////////////////////////////////////////////////////

void loadsounds(void)
{
  if (!init_xms())                sb_exist = 0;
  if (getfreexms() < XMSREQUIRED) sb_exist = 0;

#ifdef SHAREDEMB
  init_sharing();
#endif
  if (!open_sound_resource_file(SOUND_FILE)) sb_exist = 0;

  int i;
  for (i = 0; i < NUMSOUNDS; i++)
    load_sound(&(sounds[i]), sound_key[i]);

  close_sound_resource_file();
  atexit(soundexitproc);
}

//////////////////////////////////////////////////////////////////////////////

void freesounds(void)
{
  int i;
  for (i = 0; i < NUMSOUNDS; i++)
    free_sound(sounds + i);

#ifdef SHAREDEMB
  shutdown_sharing();
#endif
}

//////////////////////////////////////////////////////////////////////////////

void allocMem(void)
{
  mainbuf = (byte far *)farmalloc(64000L);
  if (mainbuf == NULL)
    errorAbort("Error: Not enough memory\n\r");

  allocImage(16, 16, &arrow);
  if (arrow.ptr == NULL)
  {
    farfree(mainbuf);
    errorAbort("Error: Not enough memory\n\r");
  }

  map = (byte far *)farmalloc((long)MAP_X_SIZE * MAP_Y_SIZE);
  if (map == NULL)
  {
    farfree(mainbuf);
    resetImage(&arrow);
    errorAbort("Error: Not enough memory\n\r");
  }

  battleMap = (byte far *)farmalloc((long)MAP_X_SIZE * MAP_Y_SIZE);
  if (battleMap == NULL)
  {
    farfree(mainbuf);
    resetImage(&arrow);
    farfree(map);
    errorAbort("Error: Not enough memory\n\r");
  }

  tiles = (tileType far *)farmalloc((long)NUM_TILES * sizeof(tileType));
  if (tiles == NULL)
  {
    farfree(mainbuf);
    resetImage(&arrow);
    farfree(map);
    errorAbort("Error: Not enough memory\n\r");
  }

  sprites = (tileType far *)farmalloc((long)NUM_TILES * sizeof(tileType));
  if (sprites == NULL)
  {
    farfree(mainbuf);
    resetImage(&arrow);
    farfree(map);
    farfree(tiles);
    errorAbort("Error: Not enough memory\n\r");
  }

  subAreaData = (subAreaType far *)farmalloc(sizeof(subAreaType) *
                                             NUM_SUB_AREAS);
  if (subAreaData == NULL)
  {
    farfree(mainbuf);
    resetImage(&arrow);
    farfree(map);
    farfree(tiles);
    farfree(sprites);
    errorAbort("Error: Not enough memory\n\r");
  }
}

//////////////////////////////////////////////////////////////////////////////

void freeMem(void)
{
  farfree(sprites);
  farfree(sprites);
  farfree(tiles);
  farfree(battleMap);
  farfree(map);
  resetImage(&arrow);
  farfree(mainbuf);
}

//////////////////////////////////////////////////////////////////////////////

void loadData(void)
{
  FILE *handle;
  long x;

  handle = fopen(CONFIG_FILE, "rb");
  if (handle != NULL)
  {
    fread(&gameconfig, 1, sizeof(gameconfig), handle);
    fclose(handle);
  }

  handle = fopen(PALETTE_FILE, "rb");
  if (handle == NULL)
    errorAbort("Error: Could not open palette file\n\r");
  fread(gamePalette, 1, 3 * 256, handle);
  fclose(handle);

  handle = fopen(IMAGE_FILE, "rb");
  if (handle == NULL)
    errorAbort("Error: Could not open image file\n\r");
  for (x = 0; x < 256; x++)
    arrow.ptr[x] = fgetc(handle);
  fclose(handle);

  loadSprites(SPRITE_FILE, (byte far *)sprites);
  loadTiles(TILES_FILE, (byte far *)tiles, walkable);

  handle = fopen(SUBAREA_FILE, "rb");
  if (handle == NULL)
    errorAbort("Error: Could not open subarea file\n\r");
  fread(subAreaData, 1, NUM_SUB_AREAS * sizeof(subAreaType), handle);

  int y;
  for (x = 0; x < NUM_MESSAGES; x++)
  {
    if (messages[x].txt_read(handle) == NULL)
    {
      fclose(handle);
      for (y = 0; y < x; y++)
        messages[x].txt_free();
      errorAbort("Error: Could not allocate message memory\n\r");
    }
  }
  fclose(handle);

  handle = fopen(GAMEDATA_FILE, "rb");
  if (handle == NULL)
    errorAbort("Error: Could not open game data file\n\r");
  fread(&gamedata, 1, sizeof(gameDataType), handle);
  fclose(handle);
}

//////////////////////////////////////////////////////////////////////////////

void Init(void)
{
  randomize();

  initKeyboard();

  allocMem();
  errorCode = ALLOC_MEM;

  loadData();
  errorCode = LOAD_DATA;

  sb_exist = 1;
  if (!detect_settings(&baseio, &irq, &dma, &dma16)) sb_exist = 0;
  if (sb_exist == 1)
    if (!init_sb(baseio, irq, dma, dma16)) sb_exist = 0;
  init_mixing();
  loadsounds();
  set_sampling_rate(11025);
  errorCode = INIT_SOUND;

  setvmode(0x13);

  getAllPalette(defaultPalette);
  setAllPalette(gamePalette);
  errorCode = SET_PALETTE;
}

//////////////////////////////////////////////////////////////////////////////

void Close(void)
{
  switch (errorCode)
  {
    SET_PALETTE:
      setAllPalette(defaultPalette);

    INIT_SOUND:
      shutdown_mixing();
      shutdown_sb();
      freesounds();

    LOAD_DATA:
      int x;
      for (x = 0; x < NUM_MESSAGES; x++)
        messages[x].txt_free();

    ALLOC_MEM:
      freeMem();
  }

  setvmode(0x3);
  closeKeyboard();
}

//////////////////////////////////////////////////////////////////////////////

int main(void)
{
  Init();
  intro();
  mainMenu();
  Close();

  FILE *handle = fopen(CONFIG_FILE, "wb");
  fwrite(&gameconfig, 1, sizeof(gameconfig), handle);
  fclose(handle);

  cprintf("The RPG v1.0\n\r"
          "Compile Date: %s\n\r"
          "Compile Time: %s\n\r"
          "\n"
          "Programmer: Chad Austin\n\r"
          "Design:     Mike Van Waardhuizen\n\r"
          "\n"
          "Special thanks to:\n\r"
          "  Ethan Brodsky for providing his freeware sound library.\n\r"
          "  All the guys on #square and #programmers (DALnet) for motivation.\n\r"
          "  SAMS publishing for the best programming books.\n\r"
          "\n"
          "Aegis Software\n\r"
          "http://www.kdsi.net/~gaustin/aegis/\n\r"
          "\n"
          "Coming soon: Castles IV\n\r",
          __DATE__,
			 __TIME__);
  HDelay(3000000);

  return 0;
}
