#include "rpg.h"

int startx, endx, ix;
int starty, endy, iy;
int pixel, numpixels;
int tileOffset;

void blitTile(int x, int y, byte tile)
{
  if (x < -19 || x > 320) return;
  if (y < -19 || y > 200) return;

  tileOffset = 0;
  if (x < 0)
  {
    startx = 0;
    tileOffset = -x;
  }
  else startx = x;

  if (y < 0)   starty = 0; else starty = y;
  if (x > 299) endx = 320; else endx = x + 20;
  if (y > 179) endy = 200; else endy = y + 20;

  numpixels = endx - startx;
  for (iy = starty; iy < endy; iy++)
    _fmemcpy(curBuf + (iy << 8) + (iy << 6) + startx,
             tiles[tile] + ((iy - y) * 20 + tileOffset),
             numpixels);
}

//////////////////////////////////////////////////////////////////////////////

void blitSprite(int x, int y, byte sprite)
{
  if (x < -19 || x > 320)  return;
  if (y < -19 || y > 200)  return;

  startx = x;    if (x < 0)   startx = 0;
  starty = y;    if (y < 0)   starty = 0;
  endx = x + 20; if (x > 299) endx = 320;
  endy = y + 20; if (y > 179) endy = 200;

  for (iy = starty; iy < endy; iy++)
    for (ix = startx; ix < endx; ix++)
    {
      pixel = sprites[sprite][(iy - y) * 20 + (ix - x)];
      if (pixel)
        curBuf[iy * 320 + ix] = pixel;
    }
}

//////////////////////////////////////////////////////////////////////////////

void loadTiles(char *file, byte far *buffer, byte walkTable[NUM_TILES])
{
  FILE *handle;
  long x;
  handle = fopen(file, "rb");
  if (handle == NULL)
    errorAbort("Error: Could not open tile file\n\r");
  for (x = 0; x < sizeof(tileType) * NUM_TILES; x++)
    buffer[x] = fgetc(handle);
  fread(walkTable, NUM_TILES, 1, handle);
  fclose(handle);
}

//////////////////////////////////////////////////////////////////////////////

void loadSprites(char *file, byte far *buffer)
{
  FILE *handle;
  long x;
  handle = fopen(file, "rb");
  if (handle == NULL)
    errorAbort("Error: Could not open sprites file\n\r");
  for (x = 0; x < sizeof(tileType) * NUM_TILES; x++)
    buffer[x] = fgetc(handle);
  fclose(handle);
}

//////////////////////////////////////////////////////////////////////////////

void loadMap(char *file, byte far *map, byte far *battleMap)
{
  FILE *handle;
  long x;
  handle = fopen(file, "rb");
  if (handle == NULL)
    errorAbort("Error: Could not open map file\n\r");
  for (x = 0; x < MAP_X_SIZE * MAP_Y_SIZE; x++)
    map[x] = fgetc(handle);
  for (x = 0; x < MAP_X_SIZE * MAP_Y_SIZE; x++)
    battleMap[x] = fgetc(handle);
  fclose(handle);
}

//////////////////////////////////////////////////////////////////////////////

void drawMap(int slideX, int slideY)
{
  register x, y;
  int tx, ty;
  for (x = -9; x < 10; x++)
    for (y = -6; y < 7; y++)
    {
      tx = x + mapx;
      if (tx < 0)              tx += MAP_X_SIZE; else
      if (tx > MAP_X_SIZE - 1) tx -= MAP_X_SIZE;
      ty = y + mapy;
      if (ty < 0)              ty += MAP_Y_SIZE; else
      if (ty > MAP_Y_SIZE - 1) ty -= MAP_Y_SIZE;
      blitTile(x * 20 + 150 + slideX, y * 20 + 90 + slideY, mapTable(tx, ty));
    }
}

//////////////////////////////////////////////////////////////////////////////

void slideMap(int x, int y)
{
  HTimer slidetimer;
  slidetimer.timerOn();

  long oldtime = slidetimer.getElapsed();

  int index, lean = 0;

  for (index = 1; index < 21; index++)
  {
    if (x == -1) { drawMap(index, 0);  direction = LEFT;  } else
    if (x == 1)  { drawMap(-index, 0); direction = RIGHT; } else
    if (y == -1) { drawMap(0, index);  direction = UP;    } else
    if (y == 1)  { drawMap(0, -index); direction = DOWN;  }
    if (index < 5 || index > 15)  lean = 0;
    else                          lean = (mapx + mapy) % 2 + 1;
    blitSprite(150, 90, direction + lean);

    if (gameconfig.vertical_sync) vsync();
    blitScreen(mainbuf);
    while (slidetimer.getElapsed() - oldtime < gameconfig.delay_rate) {}
    oldtime = slidetimer.getElapsed();
  }
  slidetimer.timerOff();
}

//////////////////////////////////////////////////////////////////////////////

void hline(int y, int x, int sx, int color)
{ clearArea(x, y, sx, 1, color); }

void vline(int x, int y, int sy, int color)
{ clearArea(x, y, 1, sy, color); }

#define bufpixel(x, y) curBuf[(y << 8) + (y << 6) + x]

#define x1 x
#define y1 y

int x2, y2;

void window(int x, int y, int sx, int sy)
{
  // ix and iy are defined at the top of this module

  x2 = x1 + sx - 1;
  y2 = y1 + sy - 1;

  // Left gradient
  for (ix = x - 4; ix < x; ix++)
    vline(ix, y, sy, 30 + (ix - x) * 2);

  // Right gradient
  for (ix = x2 + 4; ix > x2; ix--)
    vline(ix, y, sy, 30 - (ix - x2) * 2);

  // Top gradient
  for (iy = y - 4; iy < y; iy++)
    hline(iy, x, sx, 30 + (iy - y) * 2);

  // Bottom gradient
  for (iy = y2 + 4; iy > y2; iy--)
    hline(iy, x, sx, 30 - (iy - y2) * 2);

  int color;

  // Upper-left corner
  for (ix = x - 4; ix < x; ix++)
    for (iy = y - 4; iy < y; iy++)
    {
      color = 30 + (ix - x) * 2 + (iy - y) * 2;
      if (color > 18) bufpixel(ix, iy) = color;
    }

  // Upper-right corner
  for (ix = x2 + 4; ix > x2; ix--)
    for (iy = y - 4; iy < y; iy++)
    {
      color = 30 - (ix - x2) * 2 + (iy - y) * 2;
      if (color > 18) bufpixel(ix, iy) = color;
    }

  // Lower-left corner
  for (ix = x - 4; ix < x; ix++)
    for (iy = y2 + 4; iy > y2; iy--)
    {
      color = 30 + (ix - x) * 2 - (iy - y2) * 2;
      if (color > 18) bufpixel(ix, iy) = color;
    }

  // Lower-right corner
  for (ix = x2 + 4; ix > x2; ix--)
    for (iy = y2 + 4; iy > y2; iy--)
    {
      color = 30 - (ix - x2) * 2 - (iy - y2) * 2;
      if (color > 18) bufpixel(ix, iy) = color;
    }

  clearArea(x, y, sx, sy);
}
