#define byte unsigned char

typedef byte tileType[TILE_X * TILE_Y];

enum { BLANK, ITEM, WEAPON, SHIELD, HELMET, MAIL, GOLD };

struct configType
{
  long delay_rate;
  int  vertical_sync;
  int  play_sound;
};

struct messageType
{
  int length;
  char far *txt;

  messageType() { length = -1; }

  byte txt_alloc(int size);
  void txt_free(void);
  byte txt_read(FILE *handle);
  void txt_write(FILE *handle);
};

struct listType
{
  int type;
  int num;
};

struct itemType
{
  char name[13];
  int HPup;
  int MPup;
  int status;
  int cost;
};

struct weaponType
{
  char name[13];
  int attack;
  int cost;
};

struct armorType
{
  char name[13];
  int defense;
  int cost;
};

struct spellType
{
  char name[9];
  int level;
  int cost;
  int target;

  int effect;
  int factor;
};

struct gameDataType
{
  listType   items[NUM_ITEMS];

  long       levelTable[MAX_LEVEL];
  itemType   itemData[NUM_USERITEMS];
  weaponType weapons[NUM_WEAPONS];
  armorType  helmets[NUM_HELMETS];
  armorType  shields[NUM_SHIELDS];
  armorType  mails[NUM_MAILS];
  spellType  spells[NUM_SPELLS];
};

enum { PERSON, CHEST, INN, SHOP, BOSS };

struct personType
{
  int spriteSet;
  int direction;
  int lean;
  int mapx, mapy;
  int offx, offy;

  int delay;
  int speed;

  int mode;
  int curDelay;

  int message;
};

struct chestType
{
  int  spriteSet;
  int  type;
  int  mapx, mapy;
  long num;
  int  mode;

  int id;
};

struct innType
{
  int mapx, mapy;
  int spriteSet;
  char name[21];
  int cost;
};

struct shopType
{
  int      mapx, mapy;
  int      spriteSet;
  int      message;
  listType items[6];
};

struct bossType
{
  int mapx, mapy;
  int spriteSet;
  int message;
};

union EntityTypes
{
  personType person;
  chestType  chest;
  innType    inn;
  shopType   shop;
  bossType   boss;
};

struct exitType
{
  int type;
  int x, y;
  int dir;
  int lx, ly;
};

struct entityType
{
  int type;
  EntityTypes data;
};

struct subAreaType
{
  char mapFile[13];

  int startx, starty;
  int numExits;
  exitType exit[MAX_EXITS];

  int numEntities;
  entityType entity[MAX_ENTITIES];
};

struct battleType
{
  int numMonsters;
  int percent;
  int monster[16];
};

struct playerType
{
  char name[13];
  int maxHP, HP;
  int maxMP, MP;
  int status;

  int strength;
  int speed;
  int wisdom;

  long gold;
  long exp;
  int level;

  int spells[NUM_SPELLS];

  int numchests;
  byte chests[NUM_SUB_AREAS * MAX_ENTITIES];  // This shows whether a chest
                                              // has been opened
  int numitems;

  int weapon;
  int shield;
  int helmet;
  int mail;

  int attack(void);
  int mpower(void);
  int defense(void);
  int mdef(void);
};

struct monsterType
{
  char name[13];
  int maxHP, HP;
  int status;
  int attack;
  int def;
  int mdef;
  int speed;
  int gold;
  int exp;
};
